@@ -100,11 +100,14 @@ class TUISettings(BaseModel):
     displayed_events: int = 45
     max_other_fields_width: int = 90
     c_to_f: bool = True
+    flush_seconds: int = 10
+    updates_per_second: int = 2
 
 
 class EventsSettings(BaseSettings):
     verbosity: int = 0
     snaps: list[str] = []
+    scadas: list[str] = []
     paths: Paths = Paths()
     sync: SyncSettings = SyncSettings()
     mqtt: MQTTClient = MQTTClient()
