@@ -8,8 +8,11 @@ from dataclasses import field
 from datetime import datetime
 from datetime import timedelta
 from datetime import timezone
+from enum import Enum
 from pathlib import Path
 from typing import Any
+from typing import Optional
+from typing import Sequence
 
 import pandas as pd
 from anyio import to_thread
@@ -95,8 +98,22 @@ class SyncSpinners:
         return Panel(table, title="[b]Sync", border_style="blue")
 
 
+# NOTE - this is manually copied from scada code;
+#        instead it should be moved to gwproto and
+#        used from there.
+class HoneywellThermostatOperatingState(Enum):
+    idle = 0
+    heating = 1
+    pending_heat = 2
+    pending_cool = 3
+    vent_economizer = 4
+    cooling = 5
+    fan_only = 6
+
+
 class TUI:
     settings: EventsSettings
+    read_only: bool
     df: pd.DataFrame
     live_history_df: pd.DataFrame
     display_df: pd.DataFrame
@@ -111,8 +128,9 @@ class TUI:
     statuses: dict[str, GtShStatus]
     scadas_to_snap: list[str]
 
-    def __init__(self, settings: EventsSettings):
+    def __init__(self, settings: EventsSettings, read_only: bool):
         self.settings = settings
+        self.read_only = read_only
         if self.settings.paths.csv_path.exists():
             self.df = pd.read_csv(
                 self.settings.paths.csv_path,
@@ -125,9 +143,10 @@ class TUI:
                 index=pd.DatetimeIndex([], name="TimeNS"),
                 columns=["MessageId", "Src", "TypeName", "other_fields"],
             )
+            self.df.to_csv(self.settings.paths.csv_path)
         self.df.drop_duplicates("MessageId", inplace=True)
         self.live_history_df = self.df.head(0)
-        self.display_df = self.df.tail(self.settings.tui.displayed_events)
+        self.display_df = self.extract_display_df()
         self.queue = queue.Queue()
         self.gwd_text = Text()
         self.sync_spinners = SyncSpinners()
@@ -147,7 +166,14 @@ class TUI:
         for path in latest_dir.glob(f"**/*{path_suffix}"):
             with path.open() as f:
                 latest_str = f.read()
-            member[path.name[: -len(path_suffix)]] = maker.type_to_tuple(latest_str)
+            try:
+                made = maker.type_to_tuple(latest_str)
+            except Exception as e:
+                logger.exception("ERROR handling %s:\n%s\n", path, latest_str)
+                logger.exception(e)
+                # raise e
+            else:
+                member[path.name[: -len(path_suffix)]] = made
 
     def select_scadas_for_snaps(self):
         self.scadas_to_snap = []
@@ -172,6 +198,18 @@ class TUI:
     def load_snaps(self):
         self._load_latest("snap", "snaps", SnapshotSpaceheat_Maker)
 
+    def extract_display_df(self) -> pd.DataFrame:
+        if self.settings.scadas:
+            srcs_used = [
+                src
+                for src in self.df["Src"].unique()
+                if any(scada in src for scada in self.settings.scadas)
+            ]
+            filtered_df = self.df[self.df["Src"].isin(srcs_used)]
+        else:
+            filtered_df = self.df
+        return filtered_df.tail(self.settings.tui.displayed_events)
+
     def reload_dfs(self):
         self.df = pd.read_csv(
             self.settings.paths.csv_path,
@@ -181,7 +219,7 @@ class TUI:
         )
         self.df = pd.concat([self.df, self.live_history_df]).sort_index()
         self.df.drop_duplicates("MessageId", inplace=True)
-        self.display_df = self.df.tail(self.settings.tui.displayed_events)
+        self.display_df = self.extract_display_df()
 
     def make_layout(self):
         self.layout = Layout(name="root")
@@ -194,21 +232,28 @@ class TUI:
             Layout(name="latest"),
             Layout(name="events", ratio=3),
         )
-        self.layout["latest"].split(
-            Layout(name="snap0"),
-            Layout(name="snap1"),
-        )
+        # EventsSettings
+        if len(self.settings.scadas) == 1:
+            self.layout["latest"].split(
+                Layout(name="snap0"),
+            )
+        else:
+            self.layout["latest"].split(
+                Layout(name="snap0"),
+                Layout(name="snap1"),
+            )
         self.layout["footer"].split_row(
             Layout(name="GWDEvents", minimum_size=100, ratio=2),
             Layout(name="sync"),
         )
-        self.layout["header"].update(Header())
+        self.layout["header"].update(Header(self.settings.scadas))
         self.layout["events"].update(self.event_table)
         self.layout["GWDEvents"].update(
             Panel(self.gwd_text, title="[b]GWDEvents", border_style="green")
         )
         self.layout["snap0"].update(self.make_snapshot(self.scadas_to_snap[0]))
-        self.layout["snap1"].update(self.make_snapshot(self.scadas_to_snap[1]))
+        if len(self.settings.scadas) != 1:
+            self.layout["snap1"].update(self.make_snapshot(self.scadas_to_snap[1]))
         self.layout["sync"].update(self.sync_spinners.panel())
 
     def handle_gwd_event(self, event: GWDEvent) -> None:
@@ -260,9 +305,9 @@ class TUI:
             self.event_table.columns[2].no_wrap = True
             self.event_table.columns[2].max_width = 40
             self.event_table.columns[3].no_wrap = True
-            self.event_table.columns[
-                3
-            ].max_width = self.settings.tui.max_other_fields_width
+            self.event_table.columns[3].max_width = (
+                self.settings.tui.max_other_fields_width
+            )
         i = 0
         for _, row in self.display_df.tail(
             self.settings.tui.displayed_events
@@ -271,42 +316,65 @@ class TUI:
             self.add_row(row)
         return self.event_table
 
-    def handle_event(self, event: EventBase) -> None:
-        logger.debug("++handle_event")
+    def update_display(self, message_src: str, message_id: str, row_df: pd.DataFrame):
+        logger.debug("++update_display")
         path_dbg = 0
-        if event.TypeName in ["gridworks.event.problem", "gridworks.event.shutdown"]:
-            logger.info(event.json(sort_keys=True, indent=2))
-        any_event = AnyEvent(**event.dict())
-        row_df = any_event.as_dataframe(
-            columns=self.df.columns.values, interpolate_summary=True
-        )
-        display_not_full = len(self.display_df) < self.settings.tui.displayed_events
-        if not (self.display_df["MessageId"] == event.MessageId).any() and (
-            display_not_full or row_df.index[0] >= self.display_df.index[0]
+        # Check if message src is accepted
+        if not self.settings.scadas or any(
+            scada in message_src for scada in self.settings.scadas
         ):
             path_dbg |= 0x00000001
-            self.display_df = (
-                pd.concat([self.display_df, row_df])
-                .sort_index()
-                .tail(self.settings.tui.displayed_events)
-            )
-            self.layout["events"].update(self.make_event_table())
+            # Check if time is in the display window
+            if (
+                len(self.display_df) < self.settings.tui.displayed_events
+                or row_df.index[0] >= self.display_df.index[0]
+            ):
+                path_dbg |= 0x00000002
+                # Check if it is already present
+                if not (self.display_df["MessageId"] == message_id).any():
+                    path_dbg |= 0x00000004
+                    self.display_df = (
+                        pd.concat([self.display_df, row_df])
+                        .sort_index()
+                        .tail(self.settings.tui.displayed_events)
+                    )
+                    self.layout["events"].update(self.make_event_table())
+        logger.debug(f"--update_display: 0x{path_dbg:08X}")
+
+    def flush_live_history(self):
+        concatdf = pd.concat([self.df, self.live_history_df]).sort_index()
+        droppeddf = concatdf.drop_duplicates("MessageId")
+        droppeddf.to_csv(self.settings.paths.csv_path)
+        self.df = droppeddf
+        self.live_history_df = self.df.head(0)
+
+    def update_live_history(self, message_id: str, row_df: pd.DataFrame):
+        logger.debug("++update_live_history")
+        path_dbg = 0
         if (
-            not (self.live_history_df["MessageId"] == event.MessageId).any()
-            and not (self.df["MessageId"] == event.MessageId).any()
+            not self.read_only
+            and not (self.live_history_df["MessageId"] == message_id).any()
+            and not (self.df["MessageId"] == message_id).any()
         ):
-            path_dbg |= 0x00000002
+            path_dbg |= 0x00000001
             self.live_history_df = pd.concat(
                 [self.live_history_df, row_df]
             ).sort_index()
             if len(self.live_history_df) > 100:
-                path_dbg |= 0x00000004
-                concatdf = pd.concat([self.df, self.live_history_df]).sort_index()
-                droppeddf = concatdf.drop_duplicates("MessageId")
-                droppeddf.to_csv(self.settings.paths.csv_path)
-                self.df = droppeddf
-                self.live_history_df = self.df.head(0)
-        logger.debug(f"--handle_event: 0x{path_dbg:08X}")
+                path_dbg |= 0x00000002
+                self.flush_live_history()
+        logger.debug(f"--update_live_history: 0x{path_dbg:08X}")
+
+    def handle_event(self, message_src: str, event: EventBase) -> None:
+        logger.debug("++handle_event")
+        if event.TypeName in ["gridworks.event.problem", "gridworks.event.shutdown"]:
+            logger.info(event.json(sort_keys=True, indent=2))
+        row_df = AnyEvent(**event.dict()).as_dataframe(
+            columns=self.df.columns.values, interpolate_summary=True
+        )
+        self.update_display(message_src, event.MessageId, row_df)
+        self.update_live_history(event.MessageId, row_df)
+        logger.debug("--handle_event")
 
     def handle_pwr(self, pwr: GsPwr):
         pass
@@ -334,7 +402,7 @@ class TUI:
                     f.write(snap_str)
                 self.snaps[snap.FromGNodeAlias] = snap
                 self.select_scadas_for_snaps()
-                for idx in [0, 1]:
+                for idx in range(len(self.layout["latest"].children)):
                     path_dbg |= 0x00000008
                     if snap.FromGNodeAlias == self.scadas_to_snap[idx]:
                         path_dbg |= 0x00000010
@@ -373,6 +441,8 @@ class TUI:
             if (
                 telemetry_name == TelemetryName.WaterTempCTimes1000
                 or telemetry_name == TelemetryName.WaterTempCTimes1000.value
+                or telemetry_name == TelemetryName.AirTempCTimes1000
+                or telemetry_name == TelemetryName.AirTempCTimes1000.value
             ):
                 centigrade = snap.Snapshot.ValueList[i] / 1000
                 if self.settings.tui.c_to_f:
@@ -384,9 +454,30 @@ class TUI:
             elif (
                 telemetry_name == TelemetryName.WaterTempFTimes1000
                 or telemetry_name == TelemetryName.WaterTempFTimes1000.value
+                or telemetry_name == TelemetryName.AirTempFTimes1000
+                or telemetry_name == TelemetryName.AirTempFTimes1000.value
             ):
                 value_str = f"{snap.Snapshot.ValueList[i] / 1000:5.2f}"
                 unit = "F"
+            elif (
+                telemetry_name == TelemetryName.GallonsTimes100
+                or telemetry_name == TelemetryName.GallonsTimes100.value
+            ):
+                value_str = f"{snap.Snapshot.ValueList[i] / 100:5.2f}"
+                unit = "Gallons"
+            elif (
+                telemetry_name == TelemetryName.ThermostatState
+                or telemetry_name == TelemetryName.ThermostatState.value
+            ):
+                try:
+                    state_enum = HoneywellThermostatOperatingState(
+                        snap.Snapshot.ValueList[i]
+                    )
+                    enum_str = state_enum.name
+                except:  # noqa
+                    enum_str = "UNKNOWN"
+                value_str = f"{enum_str} / {snap.Snapshot.ValueList[i]}"
+                unit = "Heat State"
             else:
                 value_str = f"{snap.Snapshot.ValueList[i]}"
                 unit = snap.Snapshot.TelemetryNameList[i].value
@@ -424,7 +515,7 @@ class TUI:
                 self.handle_snapshot(message.Payload.snap)
             case EventBase():
                 path_dbg |= 0x00000020
-                self.handle_event(message.Payload)
+                self.handle_event(message.src(), message.Payload)
             case _:
                 path_dbg |= 0x00000040
         logger.debug(f"--handle_message: 0x{path_dbg:08X}")
@@ -451,25 +542,50 @@ class TUI:
             pass
 
     def loop(self):
-        with Live(self.layout, refresh_per_second=10, screen=False):
+        with Live(
+            self.layout,
+            refresh_per_second=self.settings.tui.updates_per_second,
+            screen=False,
+        ):
+            last_flush = time.time()
             while True:
-                time.sleep(0.1)
+                time.sleep(1)
                 self.check_sync_queue()
+                if not self.read_only and len(self.live_history_df) > 0:
+                    now = time.time()
+                    if now > last_flush + self.settings.tui.flush_seconds:
+                        self.flush_live_history()
+                        last_flush = now
 
     async def tui_task(self):
         await to_thread.run_sync(self.loop)
 
 
 class Header:
+    scadas: Sequence[str]
+
     """Display header with clock."""
 
+    def __init__(self, scadas: Optional[Sequence[str]] = None):
+        if scadas:
+            self.scadas = scadas[:]
+        else:
+            self.scadas = []
+
     # noinspection PyMethodMayBeStatic
     def __rich__(self) -> Panel:
         grid = Table.grid(expand=True)
         grid.add_column(justify="center", ratio=1)
         grid.add_column(justify="right")
+        title = "[b]Gridworks Events"
+        if self.scadas:
+            title += " from Scadas: "
+            for i, scada in enumerate(self.scadas):
+                title += scada
+                if i < len(self.scadas) - 1:
+                    title += ", "
         grid.add_row(
-            "[b]Gridworks Events",
+            title,
             datetime.now().ctime().replace(":", "[blink]:[/]"),
         )
         return Panel(grid, style="white on blue")
